from classic.error_handling import Error, ErrorsList


class Namespaced(Error):
    namespace = 'some_namespace'


class InheritingNamespace(Namespaced):
    pass


class NamespacedWithCode(Error):
    namespace = 'some_namespace'
    code = 'some_error'


class WithCode(Error):
    code = 'some_error'


class WithLiteralMessage(Error):
    message_template = 'some_message'


class WithMessageTemplate(Error):
    message_template = 'some_template: {arg}'


class TestError:

    def test_namespaced(self):
        error = Namespaced()

        assert error.namespace == 'some_namespace'
        assert error.code == 'namespaced'
        assert error.representation == {
            'code': 'some_namespace.namespaced',
            'context': {},
        }
        assert str(error) == '<Error "some_namespace.namespaced">'

    def test_namespaced_inheriting(self):
        error = InheritingNamespace()

        assert error.namespace == 'some_namespace'
        assert error.code == 'inheriting_namespace'
        assert error.representation == {
            'code': 'some_namespace.inheriting_namespace',
            'context': {},
        }
        assert str(error) == '<Error "some_namespace.inheriting_namespace">'

    def test_code(self):
        error = WithCode()

        assert error.namespace is None
        assert error.code == 'some_error'
        assert error.representation == {'code': 'some_error', 'context': {}}
        assert str(error) == '<Error "some_error">'

    def test_literal_message(self):
        error = WithLiteralMessage()

        assert error.namespace is None
        assert error.code == 'with_literal_message'
        assert error.representation == {
            'code': 'with_literal_message',
            'message': 'some_message',
            'context': {},
        }
        assert str(error) == '<Error "with_literal_message">'

    def test_message_template(self):
        error = WithMessageTemplate(arg=1)

        assert error.namespace is None
        assert error.code == 'with_message_template'
        assert error.representation == {
            'code': 'with_message_template',
            'message': 'some_template: 1',
            'context': {
                'arg': 1
            },
        }
        assert str(error) == '<Error "with_message_template">'


class TestErrorsList:

    def test_(self):
        errors_list = ErrorsList(WithCode(), WithCode())

        assert errors_list.code == 'errors_list'
        assert errors_list.representation == [
            {'code': 'some_error', 'context': {}},
            {'code': 'some_error', 'context': {}},
        ]



from classic.error_handling.utils import camel_case_to_dash


def test_camel_case_to_dash():
    assert camel_case_to_dash('SomeName') == 'some_name'
    assert camel_case_to_dash('some_name') == 'some_name'



from collections import defaultdict
from typing import List, ClassVar

from .utils import camel_case_to_dash


namespaces = defaultdict(list)


def represent_error(error):
    return {
        'code': error.code_representation,
        'context': error.context,
    }


def represent_error_with_message(error):
    representation = represent_error(error)
    representation['message'] = error.message
    return representation


class ErrorMeta(type):

    def __new__(mcs, name, bases, dct):
        if 'code' not in dct:
            dct['code'] = camel_case_to_dash(name)

        if 'namespace' in dct:
            assert dct['code'] not in namespaces[dct['namespace']]

            namespaces[dct['namespace']].append(dct['code'])
            namespace = dct['namespace']
        else:
            namespace = mcs.get_namescpace(mcs, bases)

        if namespace is not None:
            dct['code_representation'] = f"{namespace}.{dct['code']}"
        else:
            dct['code_representation'] = dct['code']

        if 'message_template' in dct:
            dct['representation'] = property(represent_error_with_message)
        else:
            dct['representation'] = property(represent_error)

        return super().__new__(mcs, name, bases, dct)

    def get_namescpace(mcs, classes):
        for cls in classes:
            if namespace := getattr(cls, 'namespace', None):
                return namespace


class BaseError(Exception):
    """
    Base class for application logic errors.
    """

    code: ClassVar[str] = None

    @property
    def representation(self):
        raise NotImplementedError


class Error(BaseError, metaclass=ErrorMeta):
    namespace: ClassVar[str] = None
    message_template: ClassVar[str] = None

    code_representation: str

    @property
    def representation(self):
        """Stub for syntax highlight"""
        return

    def __init__(self, **kwargs):
        self.context = kwargs
        if self.message_template:
            self.message = self.message_template.format_map(kwargs)

    def __str__(self):
        return f'<Error "{self.code_representation}">'


class ErrorsList(BaseError):
    """
    Class for situations, when application
    must send many errors in one time.
    """

    code = 'errors_list'

    def __init__(self, *args: Error):
        self.errors: List[Error] = []
        self.errors.extend(args)

    def add(self, *args):
        self.errors.extend(args)

    @property
    def representation(self):
        return [error.representation for error in self.errors]
